#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <strings.h>
#include <glib.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#define MAXTAUXI 6

int option_disque_dur = 0;

typedef unsigned char degre;

typedef struct terme_cohomologie terme_cohomologie_s, *terme_cohomologie_ref;
struct terme_cohomologie {
	degre rho,tau;
};

typedef struct terme_duale terme_duale_s, *terme_duale_ref;
typedef struct terme_duale operation_milnor_s, *operation_milnor_ref;
struct terme_duale {
	degre rho,tauh;
	degre tau[MAXTAUXI];
	degre xi[MAXTAUXI];
};
typedef struct terme_duale_dbl terme_duale_dbl_s, *terme_duale_dbl_ref;
struct terme_duale_dbl {
	terme_duale_s gauche, droite;
};

typedef GArray* steenrod_milnor;
typedef GArray* steenrod_duale;
typedef GArray* steenrod_duale_dbl;
typedef GArray* classe_cohomologie;

typedef struct memoire_psi memoire_psi_s, *memoire_psi_ref;
struct memoire_psi {
	terme_duale_s terme;
	steenrod_duale_dbl tenseur;
};
typedef struct double_sq_milnor double_sq_milnor_s, *double_sq_milnor_ref;
struct double_sq_milnor {
	int a,b;
	steenrod_milnor resultat;
};

static GArray* memoire_double_sq = NULL;
static GArray* memoire_psis = NULL;

int
binomod2(int n, int k)
{
	int a,b;
	if (k<0) return 0;
	if (k>n) return 0;
	a=n;b=k;
	while(b>0) {
		if (a%2 < b%2) return 0;
		a=a/2;
		b=b/2;
	}
	return 1;
}

int
compare_termes_duale_lexi(terme_duale_s t, terme_duale_s u)
{
	int i,d;
	/* On ignore les coefficients dans H^{*,*} */
	for (i=MAXTAUXI-1;i>=0;i--) {
		if ((d = ((int)t.tau[i] - (int)u.tau[i]))) return 0;
		if ((d = ((int)t.xi[i] - (int)u.xi[i]))) return 0;
	}
	return 0;
}

gint
compare_termes_duale_lexi_ref(gconstpointer t, gconstpointer u)
{
	return compare_termes_duale_lexi(*(terme_duale_ref)t,
			*(terme_duale_ref)u);
}

int
compare_termes_cohomologie(terme_cohomologie_s t, terme_cohomologie_s u)
{
	int d;
	if ((d = ((int)t.rho-(int)u.rho))) return d;
	if ((d = ((int)t.tau-(int)u.tau))) return d;
	return 0;
}

gint
compare_termes_cohomologie_ref(gconstpointer t, gconstpointer u)
{
	return compare_termes_cohomologie(*(terme_cohomologie_ref)t,
			*(terme_cohomologie_ref)u);
}

int
compare_termes_duale(terme_duale_s t, terme_duale_s u)
{
	unsigned int i;
	int d;
	if ((d = ((int)t.rho-(int)u.rho))) return d;
	if ((d = ((int)t.tauh-(int)u.tauh))) return d;
	for (i=0;i<MAXTAUXI;i++) {
		if ((d = ((int)t.tau[i] - (int)u.tau[i]))) return d;
		if ((d = ((int)t.xi[i] - (int)u.xi[i]))) return d;
	}
	return 0;
}

char *
filename_psi_terme_duale(terme_duale_s t)
{
	static char filename[1024];
	char tmp[80];
	unsigned int i;
	strcpy(filename, "tmp/psi");
	assert(t.rho == 0);
	assert(t.tauh == 0);
	for (i=0;i<MAXTAUXI;i++) {
		if (t.tau[i]) {
			sprintf(tmp, "-tau%d-%d", i, t.tau[i]);
			strcat(filename, tmp);
		}
		if (t.xi[i]) {
			sprintf(tmp, "-xi%d-%d", i, t.xi[i]);
			strcat(filename, tmp);
		}
	}
	return(filename);
}

gint
compare_termes_duale_ref(gconstpointer t, gconstpointer u)
{
	return compare_termes_duale(*(terme_duale_ref)t,
			*(terme_duale_ref)u);
}

int
compare_termes_duale_dbl(terme_duale_dbl_s t, terme_duale_dbl_s u)
{
	int d;
	if ((d = compare_termes_duale(t.gauche, u.gauche))) return d;
	return compare_termes_duale(t.droite, u.droite);
}

gint
compare_termes_duale_dbl_ref(gconstpointer t, gconstpointer u)
{
	return compare_termes_duale_dbl(*(terme_duale_dbl_ref)t,
			*(terme_duale_dbl_ref)u);
}



void
initialise_operation_milnor(operation_milnor_ref t)
{
	bzero(t, sizeof(operation_milnor_s));
}

void
initialise_terme_duale(terme_duale_ref  t)
{
	bzero(t, sizeof(terme_duale_s));
}

void
initialise_terme_duale_dbl(terme_duale_dbl_ref  t)
{
	bzero(t, sizeof(terme_duale_dbl_s));
}

void
print_operation_milnor(FILE *f, operation_milnor_s t)
{
	int i,x=0,y=0;
	if (t.rho) {
		if (x) fprintf(f, "*");
		x=1;
		fprintf(f, "rho");
		if (t.rho>=2) {
			fprintf(f, "^%d",t.rho);
		}
	}
	if (t.tauh) {
		if (x) fprintf(f, "*");
		x=1;
		fprintf(f, "tau");
		if (t.tauh>=2) {
			fprintf(f, "^%d",t.tauh);
		}
	}
	if (x) fprintf(f,"*");
	x=0;
	for (i=0;i<MAXTAUXI;i++) {
		if (t.tau[i]) {
			assert(t.tau[i]==1);
			x=1;
			if (i==0)
				fprintf(f, "β");
			else 
				fprintf(f, "Q_%d",i);
		}
	}
	y=0;
	for (i=0;i<MAXTAUXI;i++) {
		if (t.xi[i]) y=i;
	}
	if (!y) {
		if (!x) fprintf(f, "Id");
		return;
	}
	if (y==1) {
		fprintf(f,"P^%d",t.xi[1]);
		return ;
	}
	fprintf(f, "P(");
	for (i=1;i<=y;i++) {
		if (i>=2) fprintf(f, ",");
		fprintf(f, "%d", t.xi[i]);
	}
	fprintf(f, ",0,...)");
}

void
print_terme_duale(FILE *f, terme_duale_s t)
{
	int i,x=0;
	for (i=0;i<MAXTAUXI;i++) {
		if (t.tau[i]) {
			if (x) fprintf(f, "*");
			x=1;
			fprintf(f, "tau[%d]",i);
			if (t.tau[i]>=2) {
				fprintf(f, "^%d",t.tau[i]);
			}
		}
	}
	for (i=0;i<MAXTAUXI;i++) {
		if (t.xi[i]) {
			if (x) fprintf(f, "*");
			x=1;
			fprintf(f, "xi[%d]",i);
			if (t.xi[i]>=2) {
				fprintf(f, "^%d",t.xi[i]);
			}
		}
	}
	if (t.rho) {
		if (x) fprintf(f, "*");
		x=1;
		fprintf(f, "rho");
		if (t.rho>=2) {
			fprintf(f, "^%d",t.rho);
		}
	}
	if (t.tauh) {
		if (x) fprintf(f, "*");
		x=1;
		fprintf(f, "tau");
		if (t.tauh>=2) {
			fprintf(f, "^%d",t.tauh);
		}
	}
	if (!x) fprintf(f, "1");
}

void
print_terme_duale_dbl(FILE *f, terme_duale_dbl_s t)
{
	print_terme_duale(f, t.gauche);
	fprintf(f, "⊗");
	print_terme_duale(f, t.droite);
}

void
print_steenrod_duale_dbl(FILE *f, steenrod_duale_dbl c)
{
	unsigned int i;
	for (i=0;i<c->len;i++) {
		if (i) fprintf(f, "+");
		print_terme_duale_dbl(f, g_array_index(c, terme_duale_dbl_s, i));
	}
	if (!c->len) fprintf(f, "0");
}

void
print_steenrod_milnor(FILE *f, steenrod_milnor c)
{
	unsigned int i;
	for (i=0;i<c->len;i++) {
		if (i) fprintf(f, "+");
		print_operation_milnor(f, g_array_index(c, operation_milnor_s, i));
	}
	if (!c->len) fprintf(f, "0");
}

void
print_steenrod_duale(FILE *f, steenrod_duale c)
{
	unsigned int i;
	for (i=0;i<c->len;i++) {
		if (i) fprintf(f, "+");
		print_terme_duale(f, g_array_index(c, terme_duale_s, i));
	}
	if (!c->len) fprintf(f, "0");
}

void
simplifie_classe_cohomologie(classe_cohomologie c)
{
	unsigned int i,x;
	g_array_sort(c, compare_termes_cohomologie_ref);
	for (i=0,x=0;i<c->len;) {
		if ((i+1<c->len) && (compare_termes_cohomologie(
			g_array_index(c, terme_cohomologie_s, i),
			g_array_index(c, terme_cohomologie_s, i+1))==0)) {
			i+=2;
		} else {
			if (i!=x) {
				g_array_index(c, terme_cohomologie_s, x) =
				g_array_index(c, terme_cohomologie_s, i);
			}
			i++;x++;
		}
	}
	g_array_set_size(c, x);
}

void
simplifie_steenrod_duale(steenrod_duale c)
{
	unsigned int i,x;
	g_array_sort(c, compare_termes_duale_ref);
	for (i=0,x=0;i<c->len;) {
		if ((i+1<c->len) && (compare_termes_duale(
			g_array_index(c, terme_duale_s, i),
			g_array_index(c, terme_duale_s, i+1))==0)) {
			i+=2;
		} else {
			if (i!=x) {
				g_array_index(c, terme_duale_s, x) =
				g_array_index(c, terme_duale_s, i);
			}
			i++;x++;
		}
	}
	g_array_set_size(c, x);
}

#define simplifie_steenrod_milnor simplifie_steenrod_duale

void
simplifie_steenrod_duale_dbl(steenrod_duale_dbl c)
{
	unsigned int i,x;
	g_array_sort(c, compare_termes_duale_dbl_ref);
	for (i=0,x=0;i<c->len;) {
		if ((i+1<c->len) && (compare_termes_duale_dbl(
			g_array_index(c, terme_duale_dbl_s, i),
			g_array_index(c, terme_duale_dbl_s, i+1))==0)) {
			i+=2;
		} else {
			if (i!=x) {
				g_array_index(c, terme_duale_dbl_s, x) =
				g_array_index(c, terme_duale_dbl_s, i);
			}
			i++;x++;
		}
	}
	g_array_set_size(c, x);
}

void
ajoute_steenrod_duale(steenrod_duale dest, steenrod_duale termes)
{
	g_array_append_vals(dest, termes->data, termes->len);
	simplifie_steenrod_duale(dest);
}

void
developpe_terme_duale(steenrod_duale dest, terme_duale_s t)
{
	terme_duale_s u,t1,t2,t3;
	int k;
	g_array_set_size(dest, 0);
	for (k=0;k<MAXTAUXI;k++) {
		if (t.tau[k]>=2) {
			steenrod_duale d = g_array_new(FALSE, FALSE,
				sizeof(terme_duale_s));
			assert(k+1<MAXTAUXI);
			u=t;
			u.tau[k]-=2;
			t1=u;t2=u;t3=u;
			t1.tauh++;
			t1.xi[k+1]++;
			t2.rho++;
			t2.tau[0]++;
			t2.xi[k+1]++;
			t3.rho++;
			t3.tau[k+1]++;
			developpe_terme_duale(d, t1);
			ajoute_steenrod_duale(dest, d);
			developpe_terme_duale(d, t2);
			ajoute_steenrod_duale(dest, d);
			developpe_terme_duale(d, t3);
			ajoute_steenrod_duale(dest, d);
			g_array_free(d, TRUE);
			return ;
		}
	}
	u=t;
	u.xi[0]=0;
	g_array_append_val(dest, u);
}

void
multiplie_termes_duale(steenrod_duale dest, terme_duale_s x, terme_duale_s y)
{
	unsigned int i;
	terme_duale_s z;
	initialise_terme_duale(&z);
	z.tauh=x.tauh+y.tauh;
	z.rho=x.rho+y.rho;
	for (i=0;i<MAXTAUXI;i++) {
		z.tau[i]=x.tau[i]+y.tau[i];
		z.xi[i]=x.xi[i]+y.xi[i];
	}
	developpe_terme_duale(dest, z);
}

void
multiplie_steenrod_duale(steenrod_duale dest, steenrod_duale x,
		steenrod_duale y)
{
	unsigned int i,j;
	steenrod_duale tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	g_array_set_size(dest, 0);
	for (i=0;i<x->len;i++)
		for (j=0;j<y->len;j++) {
			multiplie_termes_duale(tmp,
				g_array_index(x, terme_duale_s, i),
				g_array_index(y, terme_duale_s, j));
			g_array_append_vals(dest, tmp->data, tmp->len);
		}
	g_array_free(tmp, TRUE);
	simplifie_steenrod_duale(dest);
}

void
multiplie_en_place_steenrod_duale(steenrod_duale dest, steenrod_duale facteur)
{
	steenrod_duale tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	multiplie_steenrod_duale(tmp, dest, facteur);
	g_array_set_size(dest, 0);
	g_array_append_vals(dest, tmp->data, tmp->len);
	g_array_free(tmp, TRUE);
}

void
calcule_lambda(steenrod_duale dest, terme_cohomologie_s terme)
/* Seulement pour t monôme de F_2[tau,rho]*/
{
	steenrod_duale lambda = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	steenrod_duale lambdatau = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	terme_duale_s u;
	unsigned int i;
	initialise_terme_duale(&u);
	u.rho = terme.rho;
	g_array_append_val(lambda, u);
	if (terme.tau) {
		terme_duale_s x;
		initialise_terme_duale(&x);
		x.tauh=1;
		g_array_append_val(lambdatau, x);
		initialise_terme_duale(&x);
		x.rho=1;
		x.tau[0]=1;
		g_array_append_val(lambdatau, x);
		simplifie_steenrod_duale(lambdatau);

		for (i=1;i<=terme.tau;i++)
			multiplie_en_place_steenrod_duale(lambda, lambdatau);
	}
	g_array_set_size(dest, 0);
	g_array_append_vals(dest, lambda->data, lambda->len);
	g_array_free(lambda, TRUE);
	g_array_free(lambdatau, TRUE);
}



void
calcule_tenseur_elem(steenrod_duale_dbl dest, terme_duale_s g, terme_duale_s d)
{
/* g_0 a tenseur d_0 b = g_0 tenseur d_0 lambda^*(a) b */
	steenrod_duale lambda = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	steenrod_duale tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	terme_duale_s g0;
	terme_duale_dbl_s z;
	unsigned int i;
	g_array_set_size(dest, 0);
	terme_cohomologie_s terme;
	terme.tau = g.tauh;
	terme.rho = g.rho;
	calcule_lambda(lambda, terme);
	initialise_terme_duale(&g0);
	for (i=0;i<MAXTAUXI;i++) {
		g0.tau[i]=g.tau[i];
		g0.xi[i]=g.xi[i];
	}
	g_array_append_val(tmp, d);
	multiplie_en_place_steenrod_duale(tmp, lambda);
	for (i=0;i<tmp->len;i++) {
		initialise_terme_duale_dbl(&z);
		z.gauche = g0;
		z.droite = g_array_index(tmp, terme_duale_s, i);
		g_array_append_val(dest, z);
	}
	g_array_free(lambda, TRUE);
	g_array_free(tmp, TRUE);
}

void
calcule_tenseur(steenrod_duale_dbl dest, steenrod_duale g, steenrod_duale d)
{
	unsigned int i,j;
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE,
		sizeof(terme_duale_dbl_s));
	g_array_set_size(dest, 0);
	for (i=0;i<g->len;i++)
		for (j=0;j<d->len;j++) {
			calcule_tenseur_elem(tmp,
				g_array_index(g, terme_duale_s, i),
				g_array_index(d, terme_duale_s, j));
			g_array_append_vals(dest, tmp->data, tmp->len);
		}
	g_array_free(tmp, TRUE);
	simplifie_steenrod_duale_dbl(dest);
}

void
multiplie_steenrod_duale_dbl(steenrod_duale_dbl dest,
	steenrod_duale_dbl x, steenrod_duale_dbl y)
{
	steenrod_duale g = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	steenrod_duale d = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
	unsigned int i,j;
	g_array_set_size(dest, 0);
	for (i=0;i<x->len;i++)
		for (j=0;j<y->len;j++) {
			multiplie_termes_duale(g,
				g_array_index(x, terme_duale_dbl_s, i).gauche,
				g_array_index(y, terme_duale_dbl_s, j).gauche);
			multiplie_termes_duale(d,
				g_array_index(x, terme_duale_dbl_s, i).droite,
				g_array_index(y, terme_duale_dbl_s, j).droite);
			calcule_tenseur(tmp, g, d);
			g_array_append_vals(dest, tmp->data, tmp->len);
		}
	g_array_free(g, TRUE);
	g_array_free(d, TRUE);
	g_array_free(tmp, TRUE);
	simplifie_steenrod_duale_dbl(dest);
}

void
multiplie_en_place_steenrod_duale_dbl(steenrod_duale_dbl dest,
		steenrod_duale_dbl facteur)
{
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
	multiplie_steenrod_duale_dbl(tmp, dest, facteur);
	g_array_set_size(dest, 0);
	g_array_append_vals(dest, tmp->data, tmp->len);
	g_array_free(tmp, TRUE);
}

void
calcule_psi_terme_internal(steenrod_duale_dbl dest, terme_duale_s t)
{
	terme_duale_dbl_s x;
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE,
		sizeof(terme_duale_dbl_s));
	steenrod_duale_dbl psi_var = g_array_new(FALSE, FALSE,
		sizeof(terme_duale_dbl_s));
	unsigned int k,i,po;
	g_array_set_size(dest, 0);
	initialise_terme_duale_dbl(&x);
	x.droite.tauh = t.tauh;
	x.droite.rho = t.rho;
	g_array_append_val(tmp, x);
	for (k=1;k<MAXTAUXI;k++) {
		if (!t.xi[k]) continue;
		g_array_set_size(psi_var, 0);
		po=1;
		for (i=0;i<=k;i++) {
			initialise_terme_duale_dbl(&x);
			if (k-i) x.droite.xi[k-i]=po;
			if (i) x.gauche.xi[i]=1;
			g_array_append_val(psi_var, x);
			po*=2;
		}
		for (i=0;i<t.xi[k];i++)
			multiplie_en_place_steenrod_duale_dbl(tmp, psi_var);
	}
	for (k=0;k<MAXTAUXI;k++) {
		if (!t.tau[k]) continue;
		g_array_set_size(psi_var, 0);
		po=1;
		for (i=0;i<=k;i++) {
			initialise_terme_duale_dbl(&x);
			if (k-i) x.droite.xi[k-i]=po;
			x.gauche.tau[i]=1;
			g_array_append_val(psi_var, x);
			po*=2;
		}
		initialise_terme_duale_dbl(&x);
		x.droite.tau[k]=1;
		g_array_append_val(psi_var, x);
		for (i=0;i<t.tau[k];i++)
			multiplie_en_place_steenrod_duale_dbl(tmp, psi_var);
	}
	g_array_append_vals(dest, tmp->data, tmp->len);
	g_array_free(tmp, TRUE);
	g_array_free(psi_var, TRUE);
	simplifie_steenrod_duale_dbl(dest);
}

void
calcule_psi_terme_memoire(steenrod_duale_dbl dest, terme_duale_s t)
{	
	if (!memoire_psis) memoire_psis = g_array_new(FALSE, FALSE, sizeof(memoire_psi_s));
	unsigned int i;
	g_array_set_size(dest, 0);
	for (i=0;i<memoire_psis->len;i++) {
		if (compare_termes_duale(t, g_array_index(memoire_psis, memoire_psi_s,
					i).terme)==0) {
			steenrod_duale_dbl res =
				g_array_index(memoire_psis, memoire_psi_s, i).tenseur;
			g_array_append_vals(dest, res->data, res->len);
			return ;
		}
	}
	calcule_psi_terme_internal(dest, t);
	memoire_psi_s m;
	m.terme = t;
	m.tenseur = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
	g_array_append_vals(m.tenseur, dest->data, dest->len);
	g_array_append_val(memoire_psis, m);
}

void
enregistre_psi(terme_duale_s t, steenrod_duale_dbl src)
{
	FILE *f = fopen(filename_psi_terme_duale(t), "w");
	assert(f != NULL);
	print_steenrod_duale_dbl(f, src);
	fprintf(f, "\n");
	fclose(f);
}

int
parse_exposant(char *src, degre *deg)
{
	char *p;
	int l;
	int d=0;
	*deg = 1;
	if (*src != '^') {
		return 0;
	}
	p=src+1;
	while((*p >= '0') && (*p <= '9')) {
		d*=10;
		d+=(*p-'0');
		p++;
	}
	l = p-src;
	if (l==1) return 0;
	*deg = d;
	return l;
}

int
parse_terme_duale(char *src, terme_duale_ref t)
{
	char *p=src;
	initialise_terme_duale(t);
	if (*p == '1') {
		return 1;
	}
	for(;;) {
		if ((p!=src)) {
			if (*p != '*') break;
			p++;
		}
		if (!strncmp(p, "xi[", 3)) {
			int i;
			p+=3;
			if (!((*p >='0') && (*p<='9'))) {
				return 0;
			}
			i = *p-'0';
			assert(i<MAXTAUXI);
			assert(i>0);
			p++;
			if (*p != ']') return 0;
			p++;
			p+=parse_exposant(p, t->xi+i);
			continue;
		}
		if (!strncmp(p, "tau[", 4)) {
			int i;
			p+=4;
			if (!((*p >='0') && (*p<='9'))) {
				return 0;
			}
			i = *p-'0';
			assert(i<MAXTAUXI);
			p++;
			if (*p != ']') return 0;
			p++;
			p+=parse_exposant(p, t->tau+i);
			continue;
		}
		if (!strncmp(p, "rho", 3)) {
			p+=3;
			p+=parse_exposant(p, &t->rho);
			continue;
		}
		if (!strncmp(p, "tau", 3)) {
			p+=3;
			p+=parse_exposant(p, &t->tauh);
			continue;
		}
	}
	return(p-src);
}

int
parse_terme_duale_dbl(char *src, terme_duale_dbl_ref t)
{
	char *p=src;
	char *tenseur = "⊗";
	int l;
	initialise_terme_duale_dbl(t);
	l = parse_terme_duale(p, &t->gauche);
	if (!l) return 0;
	p+=l;
	if (strncmp(p, tenseur, strlen(tenseur))) return 0;
	p+=strlen(tenseur);
	l = parse_terme_duale(p, &t->droite);
	if (!l) return 0;
	p+=l;
	return (p-src);
}

int
parse_steenrod_duale_dbl(steenrod_duale_dbl dest, char *src)
{
	char *p = src;
	int l;
	g_array_set_size(dest, 0);
	if (*p == '0') return 1;
	while(1) {
		terme_duale_dbl_s t;
		l = parse_terme_duale_dbl(p, &t);
		if (!l) return 0;
		g_array_append_val(dest, t);
		p+=l;
		if (!*p) return 1;
		if (*p == '\n') return 1;
		if (*p != '+') return 0;
		p++;
	}
}

int
charge_psi(steenrod_duale_dbl dest, terme_duale_s t)
{
	int fd, r;
	struct stat st;
	size_t page_size;
	size_t nbpages;
	char *p;
	fd = open(filename_psi_terme_duale(t), O_RDONLY);
	if (fd == -1) return 0;
	if (fstat(fd, &st)==-1) {
		close(fd);
		return 0;
	}
	page_size = getpagesize();
	nbpages = (st.st_size+page_size-1)/page_size;
	p = mmap(NULL, nbpages*page_size, PROT_READ, MAP_SHARED, fd, 0);
	if (p==MAP_FAILED) {
		close(fd);
		return 0;
	}
	r = parse_steenrod_duale_dbl(dest, p);
	munmap(p, nbpages*page_size);
	close(fd);
	return r;
}

void
calcule_psi_terme_disque_dur(steenrod_duale_dbl dest, terme_duale_s t)
{
	if (charge_psi(dest, t)) return;
	calcule_psi_terme_internal(dest, t);
	enregistre_psi(t, dest);
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
	assert(charge_psi(tmp, t));
	g_array_append_vals(tmp, dest->data, dest->len);
	simplifie_steenrod_duale_dbl(tmp);
	assert(tmp->len == 0);
	g_array_free(tmp, TRUE);
}

void
calcule_psi_terme(steenrod_duale_dbl dest, terme_duale_s t)
{
	if (option_disque_dur)
		calcule_psi_terme_disque_dur(dest, t);
	else
		calcule_psi_terme_memoire(dest, t);
}

void
calcule_psi(steenrod_duale_dbl dest, steenrod_duale src)
{
	unsigned int i;
	steenrod_duale_dbl tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
	g_array_set_size(dest, 0);
	for (i=0;i<src->len;i++) {
		calcule_psi_terme(tmp, g_array_index(src, terme_duale_s, i));
		g_array_append_vals(dest, tmp->data, tmp->len);
	}
	simplifie_steenrod_duale_dbl(dest);
	g_array_free(tmp, TRUE);
}

void
calcule_accouplement_elem(classe_cohomologie dest, terme_duale_s coop,
		operation_milnor_s op)
{
	terme_duale_s x,y;
	g_array_set_size(dest, 0);
	x=coop;
	y=op;
	x.rho=0;
	x.tauh=0;
	y.rho=0;
	y.tauh=0;
	if (!compare_termes_duale(x,y)) {
		terme_cohomologie_s t;
		t.tau = coop.tauh+op.tauh;
		t.rho = coop.rho+op.rho;
		g_array_append_val(dest, t);
	}
}

void
calcule_accouplement(classe_cohomologie dest, steenrod_duale coop,
		operation_milnor_s op)
{
	unsigned int i;
	classe_cohomologie tmp = g_array_new(FALSE, FALSE, sizeof(terme_cohomologie_s));
	g_array_set_size(dest, 0);
	for (i=0;i<coop->len;i++) {
		calcule_accouplement_elem(tmp, g_array_index(coop, terme_duale_s, i), op);
		g_array_append_vals(dest, tmp->data, tmp->len);
	}
	g_array_free(tmp, TRUE);
	simplifie_classe_cohomologie(dest);
}

void
calcule_accouplement_dbl_elem(classe_cohomologie dest,
		terme_duale_dbl_s tenseur,
		operation_milnor_s x, operation_milnor_s y)
{
	classe_cohomologie p = g_array_new(FALSE, FALSE, sizeof(terme_cohomologie_s));
	g_array_set_size(dest, 0);
	calcule_accouplement_elem(p, tenseur.gauche, y);
	if (!p->len) {
		g_array_free(p, TRUE);
		return ;
	}
	assert(p->len == 1);
	steenrod_duale lambda = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	calcule_lambda(lambda, g_array_index(p, terme_cohomologie_s, 0));
	steenrod_duale tmp = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	g_array_append_val(tmp, tenseur.droite);
	steenrod_duale prod = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	multiplie_steenrod_duale(prod, lambda, tmp);
	calcule_accouplement(dest, prod, x);
	g_array_free(lambda, TRUE);
	g_array_free(tmp, TRUE);
	g_array_free(prod, TRUE);
	g_array_free(p, TRUE);
}

void
calcule_accouplement_dbl(classe_cohomologie dest,
		steenrod_duale_dbl tenseur,
		steenrod_milnor x, steenrod_milnor y)
{
	classe_cohomologie tmp = g_array_new(FALSE, FALSE, sizeof(terme_cohomologie_s));
	unsigned int i,j,k;
	g_array_set_size(dest, 0);
	for (i=0;i<x->len;i++)
		for (j=0;j<y->len;j++)
			for (k=0;k<tenseur->len;k++) {
				calcule_accouplement_dbl_elem(tmp,
					g_array_index(tenseur, terme_duale_dbl_s, k),
					g_array_index(x, operation_milnor_s, i),
					g_array_index(y, operation_milnor_s, j));
				g_array_append_vals(dest, tmp->data, tmp->len);
			}
	g_array_free(tmp, TRUE);
	simplifie_classe_cohomologie(dest);
}

int
poids_terme_duale(terme_duale_s t)
{
	int po = 1;
	int w = t.tauh+t.rho;
	unsigned int k;
	for (k=0;k<MAXTAUXI;k++) {
		w+=(t.xi[k]+t.tau[k])*(po-1);
		po*=2;
	}
	return w;
}

int
poids_steenrod_milnor(steenrod_milnor op)
{
	int w=0,ww;;
	unsigned int i;
	for (i=0;i<op->len;i++) {
		ww = poids_terme_duale(g_array_index(op, terme_duale_s, i));
		if (ww>w) w=ww;
	}
	return w;
}

void
enumere_monomes_duaux_internal(steenrod_duale dest, terme_duale_s t,
	int poids, unsigned int debut)
{
	unsigned int w=0;
	unsigned int k,l;
	unsigned int expmax = 0;
	int po = 1;
	if (debut == 0) {
		terme_duale_s u = t;
		enumere_monomes_duaux_internal(dest, u, poids, 1);
		u.tau[0]=1;
		enumere_monomes_duaux_internal(dest, u, poids, 1);
		return ;
	}
	for (k=0;k<debut;k++) {
		w+=(t.xi[k]+t.tau[k])*(po-1);
		po*=2;
	}
	expmax = (poids-w)/(po-1);
	if (expmax == 0) {
		g_array_append_val(dest, t);
		return ;
	}
	assert(debut<MAXTAUXI);
	for (k=0;k<=expmax;k++)
		for(l=0;(l<=1) && (k+l<=expmax);l++) {
			terme_duale_s u = t;
			u.xi[debut]=k;
			u.tau[debut]=l;
			enumere_monomes_duaux_internal(dest, u, poids,
				debut+1);
		}
}

void
enumere_monomes_duaux(steenrod_duale dest, int poids)
{
	terme_duale_s t;
	g_array_set_size(dest, 0);
	initialise_terme_duale(&t);
	enumere_monomes_duaux_internal(dest, t, poids, 0);
	g_array_sort(dest, compare_termes_duale_lexi_ref);
}

void
decompose_sq_sur_milnor(steenrod_milnor dest, int n)
{
	operation_milnor_s t;
	initialise_operation_milnor(&t);
	t.xi[1]=n/2;
	t.tau[0]=n%2;
	g_array_append_val(dest, t);
}

void
compose_steenrod_milnor(steenrod_milnor dest, steenrod_milnor g,
	steenrod_milnor f)
{
	int poids = poids_steenrod_milnor(f)+poids_steenrod_milnor(g);
	unsigned int m,i;
	steenrod_duale monomes = g_array_new(FALSE, FALSE, sizeof(terme_duale_s));
	classe_cohomologie coeff = g_array_new(FALSE, FALSE, sizeof(terme_cohomologie_s));
	enumere_monomes_duaux(monomes, poids);
	for (m=0;m<monomes->len;m++) {
		steenrod_duale_dbl psi = g_array_new(FALSE, FALSE, sizeof(terme_duale_dbl_s));
		calcule_psi_terme(psi, g_array_index(monomes, terme_duale_s, m));
		calcule_accouplement_dbl(coeff, psi, g, f);
		g_array_free(psi, TRUE);
		for (i=0;i<coeff->len;i++) {
			terme_cohomologie_s c = g_array_index(coeff, terme_cohomologie_s, i);
			operation_milnor_s t;
			t = g_array_index(monomes, operation_milnor_s, m);
			t.rho+=c.rho;
			t.tauh+=c.tau;
			g_array_append_val(dest, t);
		}
	}
	g_array_free(monomes, TRUE);
	g_array_free(coeff, TRUE);
	simplifie_steenrod_milnor(dest);
}

steenrod_milnor
double_sq_milnor(int a, int b)
{
	unsigned int i;
	double_sq_milnor_s mem;
	if (!memoire_double_sq)
		memoire_double_sq = g_array_new(FALSE, FALSE,
				sizeof(double_sq_milnor_s));
	for (i=0;i<memoire_double_sq->len;i++) {
		mem = g_array_index(memoire_double_sq, double_sq_milnor_s, i);
		if (mem.a == a && mem.b == b) {
			return mem.resultat;
		}
	}
	mem.a = a;
	mem.b = b;
	mem.resultat = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	steenrod_milnor sq_a = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	steenrod_milnor sq_b = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	decompose_sq_sur_milnor(sq_a, a);
	decompose_sq_sur_milnor(sq_b, b);
	compose_steenrod_milnor(mem.resultat, sq_a, sq_b);
	g_array_free(sq_a, TRUE);
	g_array_free(sq_b, TRUE);
	g_array_append_val(memoire_double_sq, mem);

	return mem.resultat;
}

void
ajoute_dbl_sq_milnor(steenrod_milnor dest, int a, int b)
{
	steenrod_milnor dbl_sq = double_sq_milnor(a, b);
	g_array_append_vals(dest, dbl_sq->data, dbl_sq->len);
	simplifie_steenrod_milnor(dest);
}

void
ajoute_dbl_sq_milnor_fois_tau(steenrod_milnor dest, int a, int b)
{
	unsigned int i;
	steenrod_milnor tmp = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	ajoute_dbl_sq_milnor(tmp, a, b);
	for (i=0;i<tmp->len;i++) {
		g_array_index(tmp, operation_milnor_s, i).tauh+=1;
	}
	g_array_append_vals(dest, tmp->data, tmp->len);
	simplifie_steenrod_milnor(dest);
	g_array_free(tmp, TRUE);
}

void
ajoute_dbl_sq_milnor_fois_rho(steenrod_milnor dest, int a, int b)
{
	unsigned int i;
	steenrod_milnor tmp = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	ajoute_dbl_sq_milnor(tmp, a, b);
	for (i=0;i<tmp->len;i++) {
		g_array_index(tmp, operation_milnor_s, i).rho++;
	}
	g_array_append_vals(dest, tmp->data, tmp->len);
	simplifie_steenrod_milnor(dest);
	g_array_free(tmp, TRUE);
}

int
verifie_adem_milnor(unsigned int a, unsigned int b)
{
	unsigned int j;
	int r;
	steenrod_milnor resultatg = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	steenrod_milnor resultatd = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	steenrod_milnor difference = g_array_new(FALSE, FALSE, sizeof(operation_milnor_s));
	assert(a<2*b);
	assert(a>0);
	ajoute_dbl_sq_milnor(resultatg,a,b);
	if (a%2 && b%2) {
		for (j=1;j<=a/2;j+=2) {
			if (binomod2(b-1-j,a-2*j)) {
				ajoute_dbl_sq_milnor(resultatd, a+b-j, j);
			}
		}
	}
	if (a%2==0 && b%2==0) {
		for (j=0;j<=a/2;j++) {
			if (binomod2(b-1-j,a-2*j)) {
				if (j%2) 
					ajoute_dbl_sq_milnor_fois_tau(resultatd, a+b-j, j);
				else
					ajoute_dbl_sq_milnor(resultatd, a+b-j, j);
			}
		}
	}
	if (a%2==0 && b%2) {
		for (j=0;j<=a/2;j++) {
			if (binomod2(b-1-j,a-2*j)) {
				ajoute_dbl_sq_milnor(resultatd, a+b-j, j);
			}
		}
		for (j=1;j<=a/2;j+=2) {
			if (binomod2(b-1-j,a-2*j)) {
				ajoute_dbl_sq_milnor_fois_rho(resultatd, a+b-j-1, j);
			}
		}
	}
	if (a%2 && b%2==0) {
		for (j=0;j<=a/2;j+=2) {
			if (binomod2(b-1-j,a-2*j)) {
				ajoute_dbl_sq_milnor(resultatd, a+b-j, j);
			}
		}
		for (j=1;j<=a/2;j+=2) {
			if (binomod2(b-1-j,a-1-2*j)) {
				ajoute_dbl_sq_milnor_fois_rho(resultatd, a+b-j-1, j);
			}
		}
	}

	g_array_append_vals(difference, resultatg->data, resultatg->len);
	g_array_append_vals(difference, resultatd->data, resultatd->len);
	simplifie_steenrod_milnor(difference);
	printf("Adem(%d,%d) : ", a, b);
	print_steenrod_milnor(stdout, difference);printf("\n");
	fflush(stdout);
	r = difference->len == 0;
	g_array_free(resultatg, TRUE);
	g_array_free(resultatd, TRUE);
	g_array_free(difference, TRUE);
	return r;
}

void
efface_memoire(void)
{
	unsigned int i;
	if (memoire_double_sq) {
		for (i=0;i<memoire_double_sq->len;i++) {
			g_array_free(g_array_index(memoire_double_sq,
					double_sq_milnor_s, i).resultat, TRUE);
		}
		g_array_free(memoire_double_sq, TRUE);
		memoire_double_sq = NULL;
	}
	if (memoire_psis) {
		for (i=0;i<memoire_psis->len;i++) {
			g_array_free(g_array_index(memoire_psis,
					memoire_psi_s, i).tenseur, TRUE);
		}
		g_array_free(memoire_psis, TRUE);
		memoire_psis = NULL;
	}
}

void
verifie_adem_milnor_poids(unsigned int poids)
{
	unsigned int a,b;
	for (a=1;a<=2*poids+1;a++)
		for(b=1;b<=2*poids+1;b++) {
			if ((a/2+b/2 == poids) &&
				(a<2*b)) {
					if (!verifie_adem_milnor(a,b)) {
						printf("Les relations d'Adem sont FAUSSES en poids %d\n",
							a/2+b/2);
						fflush(stdout);
						return ;
					}
			}
		}
	printf("Les relations d'Adem sont vérifiées en poids %d\n", poids);
	fflush(stdout);
}

void
usage(void)
{
	printf("Usage: adem [ -d ] POIDS\nVerifie les coefficients des relations d'Adem motiviques en poids POIDS.\n\tUtiliser -d pour stocker des valeurs de Psi^*\n\tdans le repertoire tmp/ plutot qu'en memoire\n");
	exit(EXIT_FAILURE);
}

int
main(int argc, char **argv)
{
	int poids;
	if (argc != 2 && argc != 3) usage();
	if (argc == 3) {
		if (strcasecmp(argv[1], "-d")) usage();
		if (sscanf(argv[2], "%d", &poids)!=1) usage();
		option_disque_dur = 1;
	} else {
		if (sscanf(argv[1], "%d", &poids)!=1) usage();
	}
	assert(poids>=0);
	verifie_adem_milnor_poids(poids);
	efface_memoire();
	return 0;
}
